import numpy as np
import random
import matplotlib.pyplot as plt

#set line width
plt.rcParams['lines.linewidth'] = 6
#set general font size 
plt.rcParams['font.size'] = 12
#set font size for labels on axes
plt.rcParams['axes.labelsize'] = 18
#set size of numbers on x-axis
plt.rcParams['xtick.major.size'] = 5
#set size of numbers on y-axis
plt.rcParams['ytick.major.size'] = 5
#set size of markers
plt.rcParams['lines.markersize'] = 10

def rSquared(observed, predicted):
    error = ((predicted - observed)**2).sum()
    meanError = error/len(observed)
    return 1 - (meanError/np.var(observed))

def genNoisyParabolicData(a, b, c, xVals, fName):
    yVals = []
    for x in xVals:
        theoreticalVal = a*x**2 + b*x + c
        yVals.append(theoreticalVal + random.gauss(0, 35))
    f = open(fName,'w')
    f.write('x        y\n')
    for i in range(len(yVals)):
        f.write(str(yVals[i]) + ' ' + str(xVals[i]) + '\n')
    f.close()
 
def getPData(fileName):
    dataFile = open(fileName, 'r')
    distances = []
    masses = []
    dataFile.readline() #discard header
    for line in dataFile:
        d, m = line.split()
        distances.append(float(d))
        masses.append(float(m))
    dataFile.close()
    return (masses, distances)

def genFits(xVals, yVals, degrees):
    models = []
    for d in degrees:
        model = np.polyfit(xVals, yVals, d)
        models.append(model)
    return models

def testFits(models, degrees, xVals, yVals, title):
    plt.plot(xVals, yVals, 'o', label = 'Data')
    for i in range(len(models)):
        estYVals = np.polyval(models[i], xVals)
        error = rSquared(yVals, estYVals)
        plt.plot(xVals, estYVals,
                   label = 'Fit of degree '\
                   + str(degrees[i])\
                   + ', R2 = ' + str(round(error, 5)))
    plt.legend(loc = 'best')
    plt.title(title)

random.seed(0)

xVals = range(-10, 11, 1)
a, b, c = 3, 0, 0
genNoisyParabolicData(a, b, c, xVals, 'parabola1.txt')
genNoisyParabolicData(a, b, c, xVals, 'parabola2.txt')

degrees = (1, 2, 16)
xVals1, yVals1 = getPData('parabola1.txt')
models1 = genFits(xVals1, yVals1, degrees)
testFits(models1, degrees, xVals1, yVals1, 'Parabola 1')

plt.figure()
xVals2, yVals2 = getPData('parabola2.txt')
models2 = genFits(xVals2, yVals2, degrees)
testFits(models2, degrees, xVals2, yVals2, 'Parabola 2')
 
plt.figure()
testFits(models1, degrees, xVals2, yVals2,
        'Apply Parabola 1 Model to Parabola 2')

def splitData(xVals, yVals, fracTraining):
    trainingSize = int(len(xVals)*fracTraining)
    trainingI = random.sample(range(len(xVals)), trainingSize)
    trainingI.sort()
    trainingX, trainingY, testX, testY = [], [], [], []
    for i in range(len(xVals)):
        if i in trainingI:
            trainingX.append(xVals[i])
            trainingY.append(yVals[i])
        else:
            testX.append(xVals[i])
            testY.append(yVals[i])
    plt.figure()
    plt.plot(trainingX, trainingY, label = 'Training')
    plt.plot(testX, testY, label = 'Test')
    plt.title('Training and Test Splits')
    plt.legend()
    return (trainingX, trainingY), (testX, testY)

def fitAndValidate(xVals, yVals, degrees):
    training, test = splitData(xVals, yVals, .5)
    models = []
    for d in degrees:
        models.append(np.polyfit(training[0], training[1], d))
    for m in models:
        print([round(c, 2) for c in m])
    plt.figure()
    testFits(models, degrees, training[0], training[1],
             'Fit to Training Data')
    plt.figure()
    testFits(models, degrees, test[0], test[1],
             'Applied to Test Data')

# random.seed(0)    
# xVals = range(-20, 20, 1)
# a, b, c = 3, 0, 0
# genNoisyParabolicData(a, b, c, xVals, 'parabola.txt')
# xVals, yVals = getPData('parabola.txt')

# fitAndValidate(xVals, yVals, (2, 16))

def minkowskiDist(v1, v2, p):
    #Assumes v1 and v2 are equal length arrays of numbers
    dist = 0
    for i in range(len(v1)):
        dist += abs(v1[i] - v2[i])**p
    return dist**(1/p)

def scaleData(vals):
    #assumes vals a sequence of numbers
    vals = np.array(vals)
    mean = sum(vals)/len(vals)
    sd = np.std(vals)
    vals = vals - mean
    return vals/sd


# K-means clustering   

class Example(object):
    
    def __init__(self, name, features, label = None):
        #Assumes features is an array of floats
        self.name = name
        self.features = features
        self.label = label
        
    def dimensionality(self):
        return len(self.features)
    
    def getFeatures(self):
        return self.features[:]
    
    def getLabel(self):
        return self.label
    
    def getName(self):
        return self.name
    
    def distance(self, other):
        return minkowskiDist(self.features, other.getFeatures(), 2)
    
    def __str__(self):
        return 'f{self.name}:{self.features}:{self.label}'

class Cluster(object):
    
    def __init__(self, examples):
        """Assumes examples a non-empty list of Examples"""
        self.examples = examples
        self.centroid = self.computeCentroid()
        
    def update(self, examples):
        """Assume examples is a non-empty list of Examples
           Replace examples; return amount centroid has changed"""
        oldCentroid = self.centroid
        self.examples = examples
        self.centroid = self.computeCentroid()
        return oldCentroid.distance(self.centroid)
    
    def computeCentroid(self):
        vals = np.array([0.0]*self.examples[0].dimensionality())
        for e in self.examples: #compute mean
            vals += e.getFeatures()
        centroid = Example('centroid', vals/len(self.examples))
        return centroid

    def getCentroid(self):
        return self.centroid

    def variability(self):
        totDist = 0
        for e in self.examples:
            totDist += (e.distance(self.centroid))**2
        return totDist
        
    def members(self):
        for e in self.examples:
            yield e

    def __str__(self):
        names = []
        for e in self.examples:
            names.append(e.getName())
        names.sort()
        result = 'Cluster with centroid '\
               + str(self.centroid.getFeatures()) + ' contains:\n  '
        for e in names:
            result = result + e + ', '
        return result[:-2] #remove trailing comma and space    
        
def dissimilarity(clusters):
    """Assumes clusters a list of clusters
       Returns a measure of the total dissimilarity of the
       clusters in the list"""
    totDist = 0
    for c in clusters:
        totDist += c.variability()
    return totDist

def kmeans(examples, k, verbose = False):
    #Get k randomly chosen initial centroids, create cluster for each
    initialCentroids = random.sample(examples, k)
    clusters = []
    for e in initialCentroids:
        clusters.append(Cluster([e]))
        
    #Iterate until centroids do not change
    converged = False
    numIterations = 0
    while not converged:
        numIterations += 1
        #Create a list containing k distinct empty lists
        newClusters = []
        for i in range(k):
            newClusters.append([])
            
        #Associate each example with closest centroid
        for e in examples:
            #Find the centroid closest to e
            smallestDistance = e.distance(clusters[0].getCentroid())
            index = 0
            for i in range(1, k):
                distance = e.distance(clusters[i].getCentroid())
                if distance < smallestDistance:
                    smallestDistance = distance
                    index = i
            #Add e to the list of examples for appropriate cluster
            newClusters[index].append(e)

        for c in newClusters: #Avoid having empty clusters
            if len(c) == 0:
                raise ValueError('Empty Cluster')
        
        #Update each cluster; check if a centroid has changed
        converged = True
        for i in range(k):
            if clusters[i].update(newClusters[i]) > 0.0:
                converged = False
        if verbose:
            print('Iteration #' + str(numIterations))
            for c in clusters:
                print(c)
            print('') #add blank line
    return clusters

def trykmeans(examples, numClusters, numTrials, verbose = False):
    """Calls kmeans numTrials times and returns the result with the
          lowest dissimilarity"""
    best = kmeans(examples, numClusters, verbose)
    minDissimilarity = dissimilarity(best)
    trial = 1
    while trial < numTrials:
        try:
            clusters = kmeans(examples, numClusters, verbose)
        except ValueError:
            continue #If failed, try again
        currDissimilarity = dissimilarity(clusters)
        if currDissimilarity < minDissimilarity:
            best = clusters
            minDissimilarity = currDissimilarity
        trial += 1
    return best

# Look at cardiac data

class Patient(Example):
    pass

def getData():
    #read in data
    hrList, stElevList, ageList, prevACSList, classList = [],[],[],[],[]
    cardiacData = open('cardiacData.txt', 'r')
    for l in cardiacData:
        l = l.split(',')
        hrList.append(int(l[0]))
        stElevList.append(int(l[1]))
        ageList.append(int(l[2]))
        prevACSList.append(int(l[3]))
        classList.append(int(l[4]))
    #Build points
    points = []
    for i in range(len(hrList)):
        features = np.array([hrList[i], prevACSList[i],\
                                stElevList[i], ageList[i]])
        pIndex = str(i)
        points.append(Patient('P'+ pIndex, features, classList[i]))
    return points

# # add possibility of scaling
# def getData(toScale = False):
#     #read in data
#     hrList, stElevList, ageList, prevACSList, classList = [],[],[],[],[]
#     cardiacData = open('cardiacData.txt', 'r')
#     for l in cardiacData:
#         l = l.split(',')
#         hrList.append(int(l[0]))
#         stElevList.append(int(l[1]))
#         ageList.append(int(l[2]))
#         prevACSList.append(int(l[3]))
#         classList.append(int(l[4]))
#     if toScale:
#         hrList = scaleData(hrList)
#         stElevList = scaleData(stElevList)
#         ageList = scaleData(ageList)
#         prevACSList = scaleData(prevACSList)
#     #Build points
#     points = []
#     for i in range(len(hrList)):
#         features = np.array([hrList[i], prevACSList[i],\
#                                 stElevList[i], ageList[i]])
#         pIndex = str(i)
#         points.append(Patient('P'+ pIndex, features, classList[i]))
#     return points
    
def sensitivity(true_pos, false_neg):
    try:
        return true_pos/(true_pos + false_neg)
    except ZeroDivisionError:
        return float('nan')
    
def specificity(true_neg, false_pos):
    try:
        return true_neg/(true_neg + false_pos)
    except ZeroDivisionError:
        return float('nan')
    
def printClustering(clustering):
    """Assumes: clustering is a sequence of clusters
       Prints information about each cluster
       Returns list of fraction of pos cases in each cluster"""
    posFracs = []
    for c in clustering:
        numPts = 0
        numPos = 0
        for p in c.members():
            numPts += 1
            if p.getLabel() == 1:
                numPos += 1
        fracPos = numPos/numPts
        posFracs.append(fracPos)
        print(f'Cluster of size {numPts}, frac. pos. = {fracPos:.3f}, ' +
              f'num. pos. = {numPos}')
    return np.array(posFracs)

def testClustering(patients, numClusters, seed = 0, numTrials = 5):
    random.seed(seed)
    bestClustering = trykmeans(patients, numClusters, numTrials)
    posFracs = printClustering(bestClustering)
    return posFracs
    
# patients = getData()
# for k in (2,):
#     print('\n     Test k-means (k = ' + str(k) + ')')
#     posFracs = testClustering(patients, k, 2)

# patients = getData(True)
# for k in (2,):
#     print('\n     Test k-means (k = ' + str(k) + ')')
#     posFracs = testClustering(patients, k, 2)
    
# patients = getData(True)
# for k in (2,4,6):
#     print('\n     Test k-means (k = ' + str(k) + ')')
#     posFracs = testClustering(patients, k, 2)
